#include "td-tut-2-main.h"
#include <stdlib.h>
#include <math.h>

typedef struct GameTime
{
  float time;
  float deltaTime;
} GameTime;

GameTime gameTime = {0};

//# Enemies

#define ENEMY_MAX_COUNT 400
#define ENEMY_TYPE_NONE 0
#define ENEMY_TYPE_MINION 1

typedef struct Enemy
{
  int16_t currentX, currentY;
  int16_t nextX, nextY;
  float startMovingTime;
  uint8_t enemyType;
} Enemy;

Enemy enemies[ENEMY_MAX_COUNT];
int enemyCount = 0;

void EnemyInit()
{
  for (int i = 0; i < ENEMY_MAX_COUNT; i++)
  {
    enemies[i] = (Enemy){0};
  }
  enemyCount = 0;
}

void EnemyDraw()
{
  for (int i = 0; i < enemyCount; i++)
  {
    Enemy enemy = enemies[i];
    switch (enemy.enemyType)
    {
    case ENEMY_TYPE_MINION:
      DrawCube((Vector3){enemy.currentX, 0.2f, enemy.currentY}, 0.4f, 0.4f, 0.4f, GREEN);
      break;
    }
  }
}

void EnemyUpdate()
{
  const int16_t castleX = 0;
  const int16_t castleY = 0;

  for (int i = 0; i < enemyCount; i++)
  {
    Enemy *enemy = &enemies[i];
    if (enemy->enemyType == ENEMY_TYPE_NONE)
    {
      continue;
    }
    float speed = 1.0f;
    float transition = (gameTime.time - enemy->startMovingTime) * speed;
    if (transition >= 1.0f)
    {
      enemy->startMovingTime = gameTime.time;
      enemy->currentX = enemy->nextX;
      enemy->currentY = enemy->nextY;
      int16_t dx = castleX - enemy->currentX;
      int16_t dy = castleY - enemy->currentY;
      if (dx == 0 && dy == 0)
      {
        // enemy reached the castle; remove it
        enemy->enemyType = ENEMY_TYPE_NONE;
        continue;
      }
      if (abs(dx) > abs(dy))
      {
        enemy->nextX = enemy->currentX + (dx > 0 ? 1 : -1);
        enemy->nextY = enemy->currentY;
      }
      else
      {
        enemy->nextX = enemy->currentX;
        enemy->nextY = enemy->currentY + (dy > 0 ? 1 : -1);
      }
    }
  }
}

Enemy *EnemyTryAdd(uint8_t enemyType, int16_t currentX, int16_t currentY)
{
  Enemy *spawn = 0;
  for (int i = 0; i < enemyCount; i++)
  {
    Enemy *enemy = &enemies[i];
    if (enemy->enemyType == ENEMY_TYPE_NONE)
    {
      spawn = enemy;
      break;
    }
  }

  if (enemyCount < ENEMY_MAX_COUNT && !spawn)
  {
    spawn = &enemies[enemyCount++];
  }

  if (spawn)
  {
    spawn->currentX = currentX;
    spawn->currentY = currentY;
    spawn->nextX = currentX;
    spawn->nextY = currentY;
    spawn->enemyType = enemyType;
    spawn->startMovingTime = 0;
  }

  return spawn;
}

//# Towers

#define TOWER_MAX_COUNT 400
#define TOWER_TYPE_NONE 0
#define TOWER_TYPE_BASE 1
#define TOWER_TYPE_GUN 2

typedef struct Tower
{
  int16_t x, y;
  uint8_t towerType;
} Tower;

Tower towers[TOWER_MAX_COUNT];
int towerCount = 0;

void TowerInit()
{
  for (int i = 0; i < TOWER_MAX_COUNT; i++)
  {
    towers[i] = (Tower){0};
  }
  towerCount = 0;
}

Tower *TowerGetAt(int16_t x, int16_t y)
{
  for (int i = 0; i < towerCount; i++)
  {
    if (towers[i].x == x && towers[i].y == y)
    {
      return &towers[i];
    }
  }
  return 0;
}

Tower *TowerTryAdd(uint8_t towerType, int16_t x, int16_t y)
{
  if (towerCount >= TOWER_MAX_COUNT)
  {
    return 0;
  }

  Tower *tower = TowerGetAt(x, y);
  if (tower)
  {
    return 0;
  }

  tower = &towers[towerCount++];
  tower->x = x;
  tower->y = y;
  tower->towerType = towerType;
  return tower;
}

void TowerDraw()
{
  for (int i = 0; i < towerCount; i++)
  {
    Tower tower = towers[i];
    DrawCube((Vector3){tower.x, 0.125f, tower.y}, 1.0f, 0.25f, 1.0f, GRAY);
    switch (tower.towerType)
    {
    case TOWER_TYPE_BASE:
      DrawCube((Vector3){tower.x, 0.4f, tower.y}, 0.8f, 0.8f, 0.8f, MAROON);
      break;
    case TOWER_TYPE_GUN:
      DrawCube((Vector3){tower.x, 0.2f, tower.y}, 0.8f, 0.4f, 0.8f, DARKPURPLE);
      break;
    }
  }
}

//# Game

float nextSpawnTime = 0.0f;

void InitGame()
{
  TowerInit();
  EnemyInit();

  TowerTryAdd(TOWER_TYPE_BASE, 0, 0);
  TowerTryAdd(TOWER_TYPE_GUN, 2, 0);
  EnemyTryAdd(ENEMY_TYPE_MINION, 5, 4);
}

void GameUpdate()
{
  float dt = GetFrameTime();
  // cap maximum delta time to 0.1 seconds to prevent large time steps
  if (dt > 0.1f) dt = 0.1f;
  gameTime.time += dt;
  gameTime.deltaTime = dt;
  EnemyUpdate();

  // spawn a new enemy every second
  if (gameTime.time >= nextSpawnTime)
  {
    nextSpawnTime = gameTime.time + 1.0f;
    // add a new enemy at the boundary of the map
    int randValue = GetRandomValue(-5, 5);
    int randSide = GetRandomValue(0, 3);
    int16_t x = randSide == 0 ? -5 : randSide == 1 ? 5 : randValue;
    int16_t y = randSide == 2 ? -5 : randSide == 3 ? 5 : randValue;
    EnemyTryAdd(ENEMY_TYPE_MINION, x, y);
  }
}

int main(void)
{
  int screenWidth, screenHeight;
  GetPreferredSize(&screenWidth, &screenHeight);
  InitWindow(screenWidth, screenHeight, "Tower defense");
  SetTargetFPS(30);

  Camera3D camera = {0};
  camera.position = (Vector3){0.0f, 10.0f, 5.0f};
  camera.target = (Vector3){0.0f, 0.0f, 0.0f};
  camera.up = (Vector3){0.0f, 0.0f, -1.0f};
  camera.fovy = 45.0f;
  camera.projection = CAMERA_PERSPECTIVE;

  InitGame();

  while (!WindowShouldClose())
  {
    if (IsPaused()) {
      // canvas is not visible in browser - do nothing
      continue;
    }
    BeginDrawing();
    ClearBackground(DARKBLUE);

    BeginMode3D(camera);
    DrawGrid(10, 1.0f);
    TowerDraw();
    EnemyDraw();
    GameUpdate();
    EndMode3D();

    DrawText("Tower defense tutorial", 5, 5, 20, WHITE);
    EndDrawing();
  }

  CloseWindow();

  return 0;
}